package core.services;

import core.api.models.timemanager.tasks.repeatinginfo.DeadlineStrategy;
import core.api.models.timemanager.tasks.repeatinginfo.TaskRepeatDayInfo;
import core.api.models.timemanager.tasks.repeatinginfo.TaskRepeatWeekDayInfo;
import core.api.models.timemanager.tasks.repeatinginfo.TaskRepeatWeekInfo;
import core.entities.timemanager.*;
import core.repositories.timemanager.ScheduleRepository;
import core.repositories.timemanager.TaskRepository;
import core.repositories.timemanager.TaskSerieItemRepository;
import core.repositories.timemanager.TaskSeriesRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import util.Tupel;

import java.time.DayOfWeek;
import java.time.Duration;
import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.*;

@Service
public class TaskSeriesService {

    @Autowired private TaskRepository taskRepository;
    @Autowired private TaskSeriesRepository taskSeriesRepository;
    @Autowired private TaskSerieItemRepository taskSerieItemRepository;
    @Autowired private ScheduleRepository scheduleRepository;


    public ServiceExitCode createTaskSeries(TaskRepeatWeekInfo taskRepeatInfo) {
        HashMap<Task, Integer> offsetMap = calcWeeklyOffsetMap(taskRepeatInfo);

        TaskSerie taskSerie = new TaskSerie();
        List<Task> clonedTasks = new ArrayList<>();
        List<AbstractSchedule> clonedSchedules = new ArrayList<>();
        int weekDayIndex = 0;
        for(Map.Entry<Task, Integer> repeatingTaskInfo: offsetMap.entrySet()) {
            Task rootTask = repeatingTaskInfo.getKey();
            addRootSubTasksToTaskSerie(taskSerie, rootTask, weekDayIndex);

            int itemIndex = weekDayIndex +1;
            Tupel<Collection<Task>, Collection<AbstractSchedule>> repeatingResult = repeatTask(rootTask, taskRepeatInfo.getEndDate(), offsetMap, taskSerie, itemIndex);

            clonedTasks.addAll(repeatingResult.getValue00());
            clonedSchedules.addAll(repeatingResult.getValue01());

            weekDayIndex++;
        }


        taskSeriesRepository.save(taskSerie);
        taskRepository.saveAll(clonedTasks);
        taskSerieItemRepository.saveAll(taskSerie.getTasks());
        scheduleRepository.saveAll(clonedSchedules);

        return ServiceExitCode.OK;
    }

    private Tupel<Collection<Task>, Collection<AbstractSchedule>> repeatTask(Task rootTask, LocalDate endingDate, HashMap<Task, Integer> offsetMap,TaskSerie taskSerie, int itemIndex) {
        List<Task> clonedTasks = new ArrayList<>();
        List<AbstractSchedule> clonedSchedules = new ArrayList<>();

        LocalDate currentDate = rootTask.getStartDate().plusDays(offsetMap.get(rootTask));
        while(currentDate.isBefore(endingDate)) {
            var cloneResult = rootTask.cloneTask();
            Task clonedRootTask = cloneResult.getValue00();
            clonedTasks.addAll(cloneResult.getValue01());
            clonedSchedules.addAll(cloneResult.getValue02());

            for(Task clonedTask : cloneResult.getValue01()) {
                TaskSerieItem item = new TaskSerieItem(clonedTask, itemIndex);
                taskSerie.addItem(item);
            }

            clonedRootTask.shiftTask(offsetMap.get(rootTask));

            currentDate = currentDate.plusDays(offsetMap.get(rootTask));
            itemIndex += offsetMap.size();
        }

        return new Tupel<>(clonedTasks, clonedSchedules);
    }

    private HashMap<Task, Integer> calcWeeklyOffsetMap(TaskRepeatWeekInfo weekInfo) throws NoSuchElementException {
        HashMap<Task,Integer> offsetMap = new HashMap<>();
        weekInfo.getWeekDayInfos().sort(Comparator.comparing(TaskRepeatWeekDayInfo::getDayOfWeek));
        for(int i=0; i<weekInfo.getWeekDayInfos().size(); i++) {
            Optional<Task> requestedTask = taskRepository.findById(weekInfo.getWeekDayInfos().get(i).getTaskID());
            if(requestedTask.isEmpty()) {
                throw new NoSuchElementException();
            } else {
                int offset = weekInfo.getWeekDayInfos().get(i).getOffset()-1;
                offsetMap.put(requestedTask.get(), offset);
            }

        }
        return offsetMap;
    }

    public ServiceExitCode createTaskSeries(Task rootTask, TaskRepeatDayInfo taskRepeatInfo) {
        TaskSerie taskSerie = new TaskSerie();
        addRootSubTasksToTaskSerie(taskSerie, rootTask, 0);

        HashMap<Task, Integer> offsetMap = new HashMap<>();
        offsetMap.put(rootTask, taskRepeatInfo.getOffset());

        var repeatingResult = repeatTask(rootTask, taskRepeatInfo.getEndingDate(), offsetMap, taskSerie, 0);
        List<Task> clonedTasks = new ArrayList<>(repeatingResult.getValue00());
        List<AbstractSchedule> clonedSchedules = new ArrayList<>(repeatingResult.getValue01());


        taskSeriesRepository.save(taskSerie);
        taskRepository.saveAll(clonedTasks);
        taskSerieItemRepository.saveAll(taskSerie.getTasks());
        scheduleRepository.saveAll(clonedSchedules);

        return ServiceExitCode.OK;
    }

    private void addRootSubTasksToTaskSerie(TaskSerie taskSerie, Task rootTask, int index) {
        Queue<Task> taskQueue = new LinkedList<>(Collections.singletonList(rootTask));
        while(!taskQueue.isEmpty()) {
            Task currentTask = taskQueue.poll();

            TaskSerieItem taskSerieItem = new TaskSerieItem(currentTask, index);
            taskSerie.addItem(taskSerieItem);
        }
    }

    public void deleteTaskSeriesItem(Task task) {
        TaskSerieItem item = task.getTaskSerieItem();
        TaskSerie taskSerie = task.getTaskSerieItem().getTaskSerie();
        taskSerie.getTasks().remove(item);
        task.setTaskSerieItem(null);
        taskSerieItemRepository.delete(item);
        if(taskSerie.getTasks().isEmpty()) {
            for(TaskSerieItem taskSerieItem : taskSerie.getTasks()) {
                taskSerieItem.setTaskSerie(null);
            }
            taskSerie.getTasks().clear();
            taskSeriesRepository.delete(taskSerie);
        } else if(task.getParent() == null){
            repearIndexing(taskSerie, item.getSeriesIndex());
        }
    }

    private void repearIndexing(TaskSerie taskSerie, int deletedIndex) {
        taskSerie.getTasks().sort(Comparator.comparingInt(TaskSerieItem::getSeriesIndex));
        List<TaskSerieItem> updatedItems = new ArrayList<>();

        for(TaskSerieItem taskSerieItem : taskSerie.getTasks()) {
            if(taskSerieItem.getSeriesIndex() > deletedIndex) {
                taskSerieItem.setSeriesIndex(taskSerieItem.getSeriesIndex() -1);
                updatedItems.add(taskSerieItem);
            }
        }

        taskSerieItemRepository.saveAll(updatedItems);
    }

    public void deleteTaskSerieByTaskgroup(Taskgroup taskgroup) {
        taskSerieItemRepository.deleteByTaskgroup(taskgroup);
        taskSeriesRepository.deleteUnreferenced();
    }

    public static String convertIndexToString(int index) {
        if(index < 10) {
            return "0" + index;
        } else {
            return String.valueOf(index);
        }
    }

    public static String insertNameIndex(int seriesIndex, String taskName) {
        return taskName.replaceAll("\\$\\{i}", convertIndexToString(seriesIndex));
    }
}
